﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.IO;
using System.IO.Compression;
using System.Security.Cryptography.X509Certificates;
using System.Security.Cryptography;
using System.Xml;
using System.Security.Cryptography.Xml;
using System.Security.Cryptography.Pkcs;

namespace iBSB2BTestClient
{

     /// <summary>
    /// XML signature, encryption/decryption (CMS Enveloped) and compression/decompression class
    /// </summary>
    public class iCryptoLib
    {
        private X509Certificate2 encryptCert = null;
        private string publicCertFolder = null;
        private bool useCertFSFolder = false;
        public string ErrorMessage;
        public bool Error = false;
        //public string EncryptionOid = "2.16.840.1.101.3.4.1.42"; // AES-256 CBC
        public string EncryptionOid = "1.2.840.113549.3.7"; // TRIPLE DES - CBC
        public string encFingerPrint = null;
        public string signFingerPrint = null;
        public string encSerial = null;
        public X509Certificate2 signatureCert = null;
        public string CertStoreName = StoreName.My.ToString();
        public StoreLocation CertStoreLocation = StoreLocation.CurrentUser;
        public SymmetricAlgorithm SymAlg = new TripleDESCryptoServiceProvider();

        public bool SetEncryptionCertificate(string thumbPrint)
        {
            encryptCert = GetCert(null, thumbPrint);
            return (encryptCert != null);
        }

        public bool SetEncryptionCertificate(byte[] cert, bool save)
        {
            encryptCert = new X509Certificate2(cert);
            return (encryptCert != null);
        }

        /// <summary>
        /// Convert hex string to decimal string
        /// </summary>
        public static string HexToDecimal(string hex)
        {
            List<int> dec = new List<int> { 0 };   // decimal result

            foreach (char c in hex)
            {
                int carry = Convert.ToInt32(c.ToString(), 16);
                // initially holds decimal value of current hex digit;
                // subsequently holds carry-over for multiplication

                for (int i = 0; i < dec.Count; ++i)
                {
                    int val = dec[i] * 16 + carry;
                    dec[i] = val % 10;
                    carry = val / 10;
                }

                while (carry > 0)
                {
                    dec.Add(carry % 10);
                    carry /= 10;
                }
            }

            var chars = dec.Select(d => (char)('0' + d));
            var cArr = chars.Reverse().ToArray();
            return new string(cArr);
        }

        /// <summary>
        /// Set folder path for public Certificates stored on file system
        /// </summary>
        public string PublicCertFolder
        {
            set
            {
                publicCertFolder = value;
                if (value != null) try
                    {
                        Directory.CreateDirectory(value);
                        useCertFSFolder = true;
                    }
                    catch { }

            }
            get
            {
                return publicCertFolder;
            }
        }

        /// <summary>
        /// Encrypt data to EnvelopedCms format with AES-256 CBC
        /// </summary>
        /// <param name="fileName">Filename of the file to encrypt</param>
        /// <param name="fingerPrint">Fingerprint of the certificate to use for encryption</param>
        public byte[] EncryptData(string fileName, string fingerPrint)
        {
            byte[] data = null;
            return EncryptData(ref data, fileName, fingerPrint);
        }

        /// <summary>
        /// Encrypt data to EnvelopedCms format with AES-256 CBC
        /// </summary>
        /// <param name="data">Binary data to encrypt</param>
        /// <param name="fingerPrint">Fingerprint of the certificate to use for encryption</param>
        public byte[] EncryptData(ref byte[] data, string fingerPrint)
        {
            return EncryptData(ref data, null, fingerPrint);
        }

        /// <summary>
        /// Encrypt data to EnvelopedCms format with AES-256 CBC
        /// </summary>
        /// <param name="data">Binary data to encrypt</param>
        /// <param name="fileName">Filename of the file to encrypt</param>
        /// <param name="fingerPrint">Fingerprint of the certificate to use for encryption</param>
        public byte[] EncryptData(ref byte[] data, string fileName, string fingerPrint)
        {
            string encFilename = Path.GetTempFileName();
            bool deleteSrcFile = false;
            if (String.IsNullOrEmpty(fileName) || data != null)
            {
                fileName = Path.GetTempFileName();
                deleteSrcFile = true;
                File.WriteAllBytes(fileName, data);
                data = null;
            }

            try
            {
                if (encryptCert == null)
                    encryptCert = GetCert("", fingerPrint);
                X509Certificate2Collection recipients = new X509Certificate2Collection();
                recipients.Add(encryptCert);
                using (LargeCMS.EnvelopedLargeCMS cms = new LargeCMS.EnvelopedLargeCMS())
                {
                    cms.CertificateStoreLocation = CertStoreLocation;
                    using (FileStream fsin = new FileStream(fileName, FileMode.Open))
                    using (FileStream fsout = new FileStream(encFilename, FileMode.Create))
                    {
                        cms.EncryptAndEncode(recipients, /*LargeCMS.Win32.szOID_RSA_DES_EDE3_CBC*/EncryptionOid, fsin, fsout);
                    }
                }
                return File.ReadAllBytes(encFilename);
            }
            catch (Exception e)
            {
                ErrorMessage = "Encryption failed. Error: " + e.Message;
                Error = true;
                return null;
            }
            finally
            {
                File.Delete(encFilename);
                if (deleteSrcFile)
                    File.Delete(fileName);
            }
        }


        /// <summary>
        /// Decrypt data from EnvelopedCms format
        /// The encFingerPrint is set with certificate that data was encrypted.
        /// </summary>
        /// <param name="data">Binary data to decrypt</param>
        public byte[] DecryptData(ref byte[] data)
        {
            string encFilename = Path.GetTempFileName();
            string decfileName = Path.GetTempFileName();
            File.WriteAllBytes(encFilename, data);
            data = null;
            RunGC();
            try
            {
                using (LargeCMS.EnvelopedLargeCMS cms = new LargeCMS.EnvelopedLargeCMS())
                {
                    cms.CertificateStoreLocation = CertStoreLocation;
                    using (FileStream fsin = new FileStream(encFilename, FileMode.Open))
                    using (FileStream fsout = new FileStream(decfileName, FileMode.Create))
                    {
                        cms.DecodeAndDecrypt(fsin, fsout);
                    }
                    encSerial = cms.RecipientCertSerialNumber;
                }
                data = null;
                RunGC();
                encFingerPrint = GetCert(null, null, encSerial).Thumbprint;
                return File.ReadAllBytes(decfileName);
            }
            catch (Exception e)
            {
                ErrorMessage = "Decryption failed. Error: " + e.Message;
                Error = true;
                return null;
            }
            finally
            {
                File.Delete(encFilename);
                File.Delete(decfileName);
            }

        }

        /// <summary>
        /// Encrypt data to XMLEnc format with AES-256 CBC
        /// </summary>
        /// <param name="data">Binary data to encrypt</param>
        /// <param name="fingerPrint">Fingerprint of the certificate to use for encryption</param>

        public byte[] XMLEncryptData(ref byte[] data, string fingerPrint)
        {
            return XMLEncryptData(ref data, null, fingerPrint);
        }

        /// <summary>
        /// Encrypt data to XMLEnc format with AES-256 CBC
        /// </summary>
        /// <param name="data">Binary data to encrypt</param>
        /// <param name="fileName">Filename of the file to encrypt</param>
        /// <param name="fingerPrint">Fingerprint of the certificate to use for encryption</param>

        public byte[] XMLEncryptData(ref byte[] data, string fileName, string fingerPrint)
        {
            try
            {
                if (encryptCert == null)
                    encryptCert = GetCert("", fingerPrint);

                EncryptedXml xenc = new EncryptedXml();
                RSACryptoServiceProvider rsaKey = (RSACryptoServiceProvider)encryptCert.PublicKey.Key;
                if (SymAlg == null)
                {
                    SymAlg = new RijndaelManaged();
                    SymAlg.KeySize = 256;
                }
                try
                {
                    EncryptedXml eXml = new EncryptedXml();
                    if ((data == null) && !String.IsNullOrEmpty(fileName))
                        data = xenc.EncryptData(File.ReadAllBytes(fileName), SymAlg);
                    else
                        data = xenc.EncryptData(data, SymAlg);
                    EncryptedData edElement = new EncryptedData();
                    // Create an EncryptionMethod element so that the  
                    // receiver knows which algorithm to use for decryption.
                    if (SymAlg is TripleDES)
                    {
                        edElement.EncryptionMethod = new EncryptionMethod(EncryptedXml.XmlEncTripleDESUrl);
                    }
                    else
                    {
                        edElement.EncryptionMethod = new EncryptionMethod(EncryptedXml.XmlEncAES256Url);
                    }

                    // Encrypt the session key and add it to an EncryptedKey element.
                    EncryptedKey ek = new EncryptedKey();

                    byte[] encryptedKey = EncryptedXml.EncryptKey(SymAlg.Key, rsaKey, false);

                    ek.CipherData = new CipherData(encryptedKey);

                    ek.EncryptionMethod = new EncryptionMethod(EncryptedXml.XmlEncRSA15Url);

                    // Set the KeyInfo element to specify the 
                    // name of the RSA key. 

                    // Create a new KeyInfo element.
                    edElement.KeyInfo = new KeyInfo();

                    KeyInfoX509Data keyInfoData = new KeyInfoX509Data(encryptCert);

                    // Add the KeyInfoName element to the  
                    // EncryptedKey object.
                    ek.KeyInfo.AddClause(keyInfoData);

                    // Add the encrypted key to the  
                    // EncryptedData object.

                    edElement.KeyInfo.AddClause(new KeyInfoEncryptedKey(ek));

                    // Add the encrypted element data to the  
                    // EncryptedData object.
                    edElement.CipherData.CipherValue = data;
                    data = null;
                    //MemoryStream ms = new MemoryStream();
                    //edElement.GetXml().OwnerDocument.Save(ms);
                    //edElement = null;
                    RunGC();
                    //return ms.ToArray();
                    //XmlWriter wr = XmlWriter.Create("");
                    return System.Text.Encoding.UTF8.GetBytes(edElement.GetXml().OuterXml);
                }
                finally
                {
                    SymAlg.Clear();
                }
            }
            catch (Exception e)
            {
                ErrorMessage = "XML Encryption failed. Error: " + e.Message;
                Error = true;
                return null;
            }
        }

        /// <summary>
        /// Decrypt data from XMLEnc format
        /// The encFingerPrint is set with certificate that data was encrypted.
        /// </summary>
        /// <param name="data">Binary data to decrypt</param>
        public byte[] XMLDecryptData(ref byte[] data)
        {
            try
            {
                XmlDocument Doc = new XmlDocument();
                Doc.Load(new MemoryStream(data));
                data = null;
                RunGC();
                EncryptedXml exml = new EncryptedXml();
                XmlNamespaceManager nsm = new XmlNamespaceManager(Doc.NameTable);
                nsm.AddNamespace("enc", EncryptedXml.XmlEncNamespaceUrl);
                XmlNodeList encryptedDataList = Doc.SelectNodes("//enc:EncryptedData", nsm);
                if (encryptedDataList != null)
                {
                    foreach (XmlNode encryptedDataNode in encryptedDataList)
                    {
                        XmlElement encryptedDataElement = encryptedDataNode as XmlElement;
                        EncryptedData ed = new EncryptedData();
                        EncryptedKey ek = new EncryptedKey();
                        XmlElement encryptedKeyDataElement = (XmlElement)encryptedDataElement.SelectSingleNode("//enc:EncryptedKey", nsm);
                        if (encryptedKeyDataElement != null)
                        {
                            ek.LoadXml(encryptedKeyDataElement);
                            var x509data = ek.KeyInfo.OfType<KeyInfoX509Data>().First();
                            if (x509data != null)
                            {
                                encryptCert = x509data.Certificates[0] as X509Certificate2;
                            }
                        }
                        ed.LoadXml(encryptedDataElement);
                        SymmetricAlgorithm symAlg = exml.GetDecryptionKey(ed, null);
                        if (symAlg == null)
                            throw new CryptographicException("Cryptography_Xml_MissingDecryptionKey");
                        return exml.DecryptData(ed, symAlg);
                        //ReplaceData(encryptedDataElement, decrypted);
                    }
                }
                return null;
            }
            catch (Exception e)
            {
                ErrorMessage = "XML Decryption failed. Error: " + e.Message;
                Error = true;
                return null;
            }
        }
  
        /// <summary>
        /// Create enveloped signature and return signed XML
        /// </summary>
        /// <param name="data">Binary data to sign</param>
        /// <param name="URIString">URIString do not use it!</param>
        /// <param name="fingerPrint">FingerPrint of the certificate to use for the signature</param>
        public byte[] SignXmlData(byte[] data, string URIString, string fingerPrint)
        {
            try
            {
                X509Certificate2 cert = GetCert("", fingerPrint);
                // Get public key
                RSACryptoServiceProvider key = (RSACryptoServiceProvider)cert.PrivateKey;

                // Create a new XML document.
                XmlDocument doc = new XmlDocument();
                // Keep white spaces.
                doc.PreserveWhitespace = true;

                doc.Load(new MemoryStream(data));

                // Create a SignedXml object.
                SignedXml signedXml = new SignedXml(doc);

                // Assign the key to the SignedXml object.
                signedXml.SigningKey = key;

                // Create a reference to be signed.
                Reference reference = new Reference();

                // Add the passed URI to the reference object.
                reference.Uri = URIString;

                XmlDsigEnvelopedSignatureTransform env = new XmlDsigEnvelopedSignatureTransform();
                reference.AddTransform(env);

                // Add the reference to the SignedXml object.
                signedXml.AddReference(reference);

                // Add an RSAKeyValue KeyInfo (optional; helps recipient find key to validate).
                KeyInfo keyInfo = new KeyInfo();
                KeyInfoX509Data keyInfoData = new KeyInfoX509Data(cert);
                keyInfo.AddClause(keyInfoData);
                signedXml.KeyInfo = keyInfo;

                // Compute the signature.
                signedXml.ComputeSignature();

                // Get the XML representation of the signature and save 
                // it to an XmlElement object.
                XmlElement xmlDigitalSignature = signedXml.GetXml();

                // Append the element to the XML document.
                doc.DocumentElement.AppendChild(doc.ImportNode(xmlDigitalSignature, true));

                if (doc.FirstChild is XmlDeclaration)
                {
                    doc.RemoveChild(doc.FirstChild);
                }


                using (var ms = new MemoryStream())
                {
                    doc.Save(ms);
                    doc = null;
                    RunGC();
                    return ms.ToArray();
                }
            }
            catch (Exception e)
            {
                ErrorMessage = "XML signature failed. Error: " + e.Message;
                Error = true;
                return null;
            }

        }

        /// <summary>
        /// Create enveloped signature with signing time and return signed XML
        /// </summary>
        /// <param name="data">Binary data to sign</param>
        /// <param name="URIString">URIString do not use it!</param>
        /// <param name="fingerPrint">FingerPrint of the certificate to use for the signature</param>
        public byte[] SignXmlDataST(ref byte[] data, string URIString, string fingerPrint)
        {
            return SignXmlDataST(ref data, null, URIString, fingerPrint);
        }

        /// <summary>
        /// Create enveloped signature with signing time and return signed XML
        /// </summary>
        /// <param name="fileName">Filename of the file to sign</param>
        /// <param name="URIString">URIString do not use it!</param>
        /// <param name="fingerPrint">FingerPrint of the certificate to use for the signature</param>
        public byte[] SignXmlDataST(string fileName, string URIString, string fingerPrint)
        {
            byte[] data = null;
            return SignXmlDataST(ref data, fileName, URIString, fingerPrint);
        }

        /// <summary>
        /// Create enveloped signature with signing time and return signed XML
        /// </summary>
        /// <param name="data">Binary data to sign</param>
        /// <param name="fileName">Filename of the file to sign, it's used if data parameter is null</param>
        /// <param name="URIString">URIString do not use it!</param>
        /// <param name="fingerPrint">FingerPrint of the certificate to use for the signature</param>
        public byte[] SignXmlDataST(ref byte[] data, string fileName, string URIString, string fingerPrint)
        {
            try
            {
                X509Certificate2 cert = GetCert("", fingerPrint);
                // Get public key
                RSACryptoServiceProvider key = (RSACryptoServiceProvider)cert.PrivateKey;

                // Create a new XML document.
                XmlDocument doc = new XmlDocument();
                // Keep white spaces.
                doc.PreserveWhitespace = true;
                if ((data == null) && !String.IsNullOrEmpty(fileName))
                    doc.Load(fileName);
                else
                    doc.Load(new MemoryStream(data));
                data = null;
                RunGC();
                // Create a SignedXml object.
                //SignedXml signedXml = new SignedXml(doc);
                URIString = "";
                SignaturePropertiesSignedXml signedXml = new SignaturePropertiesSignedXml(doc, "", "");

                // Assign the key to the SignedXml object.
                signedXml.SigningKey = key;


                // create a timestamp property
                XmlElement timestamp = doc.CreateElement("SigningTime", signedXml.XmlXAdESTSNamespaceUrl);
                timestamp.InnerText = DateTime.Now.ToUniversalTime().ToString("u");
                signedXml.AddProperty(timestamp);
                signedXml.AddProperty(signedXml.GetSigningCertificateNode(cert));

                // Create a reference to be signed.
                Reference reference = new Reference();

                // Add the passed URI to the reference object.
                reference.Uri = URIString;

                XmlDsigEnvelopedSignatureTransform env = new XmlDsigEnvelopedSignatureTransform();
                reference.AddTransform(env);

                // Add the reference to the SignedXml object.
                signedXml.AddReference(reference);

                // Add an RSAKeyValue KeyInfo (optional; helps recipient find key to validate).
                KeyInfo keyInfo = new KeyInfo();
                KeyInfoX509Data keyInfoData = new KeyInfoX509Data(cert);
                keyInfo.AddClause(keyInfoData);
                signedXml.KeyInfo = keyInfo;

                // Compute the signature.
                signedXml.ComputeSignature();

                // Get the XML representation of the signature and save 
                // it to an XmlElement object.
                XmlElement xmlDigitalSignature = signedXml.GetXml();

                // Append the element to the XML document.
                doc.DocumentElement.AppendChild(doc.ImportNode(xmlDigitalSignature, true));

                if (doc.FirstChild is XmlDeclaration)
                {
                    doc.RemoveChild(doc.FirstChild);
                }


                using (var ms = new MemoryStream())
                {
                    doc.Save(ms);
                    doc = null;
                    RunGC();
                    return ms.ToArray();
                }
            }
            catch (Exception e)
            {
                ErrorMessage = "XML signature failed. Error: " + e.Message;
                Error = true;
                return null;
            }

        }

        public void RunGC()
        {
            GC.Collect();
            GC.WaitForPendingFinalizers();
            GC.Collect();
        }

        /// <summary>
        /// Verify XML signature
        /// </summary>
        /// <param name="data">Binary data to verify</param>
        public bool VerifyXmlSignature(byte[] data)
        {
            try
            {
                // Create a new XML document.
                XmlDocument xmlDocument = new XmlDocument();

                // Format using white spaces.
                xmlDocument.PreserveWhitespace = true;

                // Load the passed XML file into the document. 
                try
                {
                    xmlDocument.Load(new MemoryStream(data));
                }
                catch (Exception e)
                {
                    throw new CryptographicException("Unable to load XML document. Error: " + e.Message);
                }
                data = null;
                RunGC();
                // Create a new SignedXml object and pass it 
                // the XML document class.
                SignedXml signedXml = new SignedXml(xmlDocument);

                // Find the "Signature" node and create a new 
                // XmlNodeList object.
                XmlNodeList nodeList = xmlDocument.GetElementsByTagName("Signature", SignedXml.XmlDsigNamespaceUrl);

                // Throw an exception if no signature was found.
                if (nodeList.Count <= 0)
                {
                    throw new CryptographicException("No Signature was found in the document.");
                }

                // This example only supports one signature for
                // the entire XML document.  Throw an exception
                // if more than one signature was found.
                if (nodeList.Count >= 2)
                {
                    throw new CryptographicException("More that one signature was found for the document.");
                }

                // Load the signature node.
                signedXml.LoadXml((XmlElement)nodeList[0]);

                var x509data = signedXml.Signature.KeyInfo.OfType<KeyInfoX509Data>().First();
                if (x509data != null)
                {
                    signatureCert = x509data.Certificates[0] as X509Certificate2;
                }
                if (signatureCert == null)
                {
                    ErrorMessage = "Signing certficate not present in the XML.";
                    return false;
                }
                signFingerPrint = signatureCert.Thumbprint;
                // Check the signature and return the result. 
                bool ok = signedXml.CheckSignature(signatureCert, false);
                signedXml = null;
                RunGC();
                return ok;
            }
            catch (Exception ex)
            {
                ErrorMessage = "Verification of XML data failed: " + ex.Message;
                Error = true;
                return false;
            }

        }
   
        public X509Certificate2 GetCert(string certSubject, string thumbPrint)
        {
            if (useCertFSFolder)
                return GetCertFromFolder(thumbPrint);
            else
                return GetCertFromStore(certSubject, thumbPrint);
        }

        public X509Certificate2 GetCert(string certSubject, string thumbPrint, string certSerial)
        {
            if (useCertFSFolder)
                return GetCertFromFolder(thumbPrint, certSerial);
            else
                return GetCertFromStore(certSubject, thumbPrint, certSerial);
        }

        public X509Certificate2 GetCertFromFolder(string thumbPrint)
        {
            return new X509Certificate2("");
        }

        public X509Certificate2 GetCertFromFolder(string thumbPrint, string certSerial)
        {          
            return new X509Certificate2("");
        }

        public X509Certificate2 GetCertFromStore(string certSubject, string thumbPrint)
        {
            return GetCertFromStore(certSubject, thumbPrint, null);
        }

        public X509Certificate2 GetCertFromStore(string certSubject, string thumbPrint, string certSerial)
        {
            X509Store my = new X509Store(CertStoreName, CertStoreLocation);
            my.Open(OpenFlags.ReadOnly);
            X509Certificate2 fcert = null;
            string certs = "";
            if (!String.IsNullOrEmpty(thumbPrint))
                thumbPrint = thumbPrint.Trim().Replace(" ", "").ToUpper();
            foreach (X509Certificate2 cert in my.Certificates)
            {
                certs += "\t" + cert.Subject + ", t=" + cert.Thumbprint + Environment.NewLine;
                if (!String.IsNullOrEmpty(certSubject) && cert.Subject.Contains(certSubject))
                {
                    if (!String.IsNullOrEmpty(thumbPrint) && !cert.Thumbprint.Equals(thumbPrint))
                    {
                        continue;
                    }
                    fcert = cert;
                    break;
                }
                else if (!String.IsNullOrEmpty(thumbPrint) && cert.Thumbprint.Equals(thumbPrint))
                {
                    fcert = cert;
                    break;
                }
                else if (!String.IsNullOrEmpty(certSerial) && cert.SerialNumber.Equals(certSerial))
                {
                    fcert = cert;
                    break;
                }
            }
            if (fcert == null)
            {
                //throw new Exception("Valid certificate '" + certSubject + (!String.IsNullOrEmpty(thumbPrint) ? "(" + thumbPrint + ")" : "") + "' was not found, among: " + Environment.NewLine + certs);
                throw new Exception("Valid certificate '" + certSubject + (!String.IsNullOrEmpty(thumbPrint) ? "(" + thumbPrint + ")" : "") + "' not found.");
            }
            return fcert;
        }

        /// <summary>
        /// Compress data with DeflateStream
        /// </summary>
        /// <param name="data">Filename of the file to zip</param>
        public byte[] ZipData(string fileName)
        {
            try
            {
                using (FileStream dataMS = new FileStream(fileName, FileMode.Open))
                using (MemoryStream compressedMS = new MemoryStream())
                using (DeflateStream ds = new DeflateStream(compressedMS, CompressionMode.Compress))
                {
                    dataMS.CopyTo(ds);
                    ds.Close();
                    return compressedMS.ToArray();
                }
            }
            catch (Exception e)
            {
                Error = true;
                ErrorMessage = "Compression failed: " + e.Message;
                return null;
            }
            finally
            {
                RunGC();
            }
        }

        /// <summary>
        /// Compress data with DeflateStream
        /// </summary>
        /// <param name="data">Binary data to zip</param>
        public byte[] ZipData(ref byte[] data)
        {
            try
            {
                using (MemoryStream dataMS = new MemoryStream(data))
                using (MemoryStream compressedMS = new MemoryStream())
                using (DeflateStream ds = new DeflateStream(compressedMS, CompressionMode.Compress))
                {
                    data = null;
                    RunGC();
                    dataMS.CopyTo(ds);
                    ds.Close();
                    return compressedMS.ToArray();
                }
            }
            catch (Exception e)
            {
                Error = true;
                ErrorMessage = "Compression failed: " + e.Message;
                return null;
            }
            finally
            {
                RunGC();
            }
        }

        /// <summary>
        /// Decompress data with DeflateStream
        /// </summary>
        /// <param name="data">Binary data to unzip</param>
        public byte[] UnZipData(ref byte[] data)
        {
            try
            {
                using (MemoryStream dataMS = new MemoryStream(data))
                using (MemoryStream output = new MemoryStream())
                using (DeflateStream ds = new DeflateStream(dataMS, CompressionMode.Decompress))
                {
                    data = null;
                    RunGC();
                    ds.CopyTo(output);
                    ds.Close();
                    return output.ToArray();
                }
            }
            catch (Exception e)
            {
                Error = true;
                ErrorMessage = "Decompression failed: " + e.Message;
                return null;
            }
            finally
            {
                RunGC();
            }
        }

#if !NOPDF

        /// <summary>
        /// Sign PDF data with SignPDF library
        /// </summary>
        /// <param name="data">Binary data of PDF to sign</param>
        public byte[] SignPDF(ref byte[] data, string certFP)
        {
            return SignPDF(ref data, null, certFP);
        }

        /// <summary>
        /// Sign PDF data with SignPDF library
        /// </summary>
        /// <param name="fileName">File name of PDF to sign</param>
        public byte[] SignPDF(string fileName, string certFP)
        {
            byte[] data = null;
            return SignPDF(ref data, fileName, certFP);
        }

        /// <summary>
        /// Sign PDF data with SignPDF library
        /// </summary>
        /// <param name="data">Binary data of PDF to sign</param>
        /// <param name="fileName">File name of PDF to sign</param>
        public byte[] SignPDF(ref byte[] data, string fileName, string certFP)
        {
            try
            {
                // uporaba komponente za podpisovanje PDF-ov
                return null;
            }
            catch (Exception e)
            {
                Error = true;
                ErrorMessage = "PDF signature failed: " + e.Message;
                return null;
            }
            finally
            {
                RunGC();
            }
        }

        /// <summary>
        /// Verify digital signature of PDF file with SignPDF library
        /// </summary>
        /// <param name="data">Binary data of PDF to verify signature</param>
        public bool VerifyPDF(byte[] data)
        {
            // uporaba komponente za podpisovanje PDF-ov
            return true;
        }
#endif

#if !NOZIP
        /// <summary>
        /// Compress data to zip file with DotNetZip library
        /// </summary>
        /// <param name="data">Binary data to zip</param>
        /// <param name="pkgfileName">File name to use for the file in zip</param>
        public byte[] ZipDataToZipFile(ref byte[] data, string pkgfileName)
        {
            return ZipDataToZipFile(ref data, null, pkgfileName);
        }

        /// <summary>
        /// Compress data to zip file with DotNetZip library
        /// </summary>
        /// <param name="fileName">File name of the file to zip</param>
        public byte[] ZipDataToZipFile(string fileName)
        {
            byte[] data = null;
            return ZipDataToZipFile(ref data, fileName, Path.GetFileName(fileName));
        }

        /// <summary>
        /// Compress data to zip file with DotNetZip library
        /// </summary>
        /// <param name="data">Binary data to zip</param>
        /// <param name="fileName">File name of the file to zip</param>
        /// <param name="pkgfileName">File name to use for the file in zip</param>
        public byte[] ZipDataToZipFile(ref byte[] data, string fileName, string pkgfileName)
        {
            if (String.IsNullOrEmpty(fileName)) fileName = "data";
            try
            {
                using (Ionic.Zip.ZipFile zip = new Ionic.Zip.ZipFile())
                using (MemoryStream ms = new MemoryStream())
                {
                    if ((data == null) && !String.IsNullOrEmpty(fileName))
                        zip.AddEntry(pkgfileName, fileName);
                    else
                        zip.AddEntry(pkgfileName, data);
                    zip.Save(ms);
                    data = null;
                    RunGC();
                    return ms.ToArray();
                }
            }
            catch (Exception e)
            {
                Error = true;
                ErrorMessage = "Compression failed: " + e.Message;
                return null;
            }
            finally
            {
                RunGC();
            }
        }

        /// <summary>
        /// Decompress data from zip file with DotNetZip library
        /// </summary>
        /// <param name="data">Binary data to unzip</param>
        public byte[] UnZipDataFromZipFile(ref byte[] data)
        {
            try
            {
                using (MemoryStream ms = new MemoryStream(data))
                using (Ionic.Zip.ZipFile zip = Ionic.Zip.ZipFile.Read(ms))
                using (var mse = new MemoryStream())
                {
                    data = null;
                    RunGC();
                    if (zip.Entries.Count != 1)
                    {
                        Error = true;
                        ErrorMessage = String.Format("Decompression failed: Zip package contains {0} files. Only packages with one file allowed.", zip.Entries.Count);
                        return null;
                    }
                    foreach (var ze in zip.Entries)
                    {
                        ze.Extract(mse);
                        break;
                    }
                    return mse.ToArray();
                }

            }
            catch (Exception e)
            {
                Error = true;
                ErrorMessage = "Decompression failed: " + e.Message;
                return null;
            }
            finally
            {
                RunGC();
            }
        }

#endif

        /// <summary>
        /// Copy stream to another stream, for .NET 3.5.
        /// </summary>
        /// <param name="input"></param>
        /// <param name="output"></param>
        public void CopyStream(Stream input, Stream output)
        {
            byte[] buffer = new byte[4096];
            int read;
            while ((read = input.Read(buffer, 0, buffer.Length)) > 0)
            {
                output.Write(buffer, 0, read);
            }
        }
    }

    /// <summary>
    /// XML signature class which enables SignatureProperties
    /// </summary>   
    public sealed class SignaturePropertiesSignedXml : SignedXml
    {
        private XmlDocument doc = null;
        private XmlElement signaturePropertiesRoot = null;
        private XmlElement signedSigProperties = null;
        private XmlElement signedProperties = null;
        public string XmlXAdESTSNamespaceUrl = "http://uri.etsi.org/01903/v1.1.1#";

        /// <summary>
        /// Create a SignedXml class which can have SignatureProperties
        /// </summary>
        /// <param name="doc">XML document the signature belongs in</param>
        /// <param name="signatureId">ID of the signature to create</param>
        /// <param name="propertiesId">ID of the SignatureProperties to create</param>
        public SignaturePropertiesSignedXml(XmlDocument doc, string signatureId, string propertiesId)
            : base(doc)
        {
            /*
            if (String.IsNullOrEmpty(signatureId))
                throw new ArgumentException("signatureId cannot be empty", "signatureId");
            if (String.IsNullOrEmpty(propertiesId))
                throw new ArgumentException("propertiesId cannot be empty", "propertiesId");
            */
            this.doc = doc;
            if (String.IsNullOrEmpty(propertiesId))
                propertiesId = "SignedProperties-" + Guid.NewGuid().ToString().ToUpper();
            if (String.IsNullOrEmpty(signatureId))
                signatureId = "Signature-" + Guid.NewGuid().ToString().ToUpper();
            Signature.Id = signatureId;

            // create a root element to hold the properties
            signaturePropertiesRoot = doc.CreateElement("QualifyingProperties", XmlXAdESTSNamespaceUrl);
            signaturePropertiesRoot.SetAttribute("Target", "#" + signatureId);
            //signaturePropertiesRoot.SetAttribute("Id", propertiesId);

            signedProperties = doc.CreateElement("SignedProperties", XmlXAdESTSNamespaceUrl);
            signedProperties.SetAttribute("Id", propertiesId);
            signaturePropertiesRoot.AppendChild(signedProperties);

            signedSigProperties = doc.CreateElement("SignedSignatureProperties", XmlXAdESTSNamespaceUrl);
            signedProperties.AppendChild(signedSigProperties);

            // create a data object for the properties to go into
            System.Security.Cryptography.Xml.DataObject signatureProperties = new System.Security.Cryptography.Xml.DataObject();
            signatureProperties.Data = signaturePropertiesRoot.SelectNodes(".");
            AddObject(signatureProperties);

            // and add a reference to the data object
            Reference propertiesRef = new Reference("#" + propertiesId);
            propertiesRef.Type = XmlXAdESTSNamespaceUrl + "SignedProperties";
            // to je potrebno zato, da potem zna sam sebe preveriti
            // problem so namespacei, ki jih ima sam xml in se ne prenesejo v SignedInfo
            // ce pa je spodnji transform, pa to niti ni potrebno in potem se izognemo problemu
            propertiesRef.AddTransform(new XmlDsigExcC14NTransform());
            AddReference(propertiesRef);

            return;
        }

        /// <summary>
        /// Returns <SigningCertificate> of the XAdES schema
        /// </summary>
        /// <param name="cert">X509Certificate2 to create the SigningCertificate node</param>
        public XmlElement GetSigningCertificateNode(X509Certificate2 cert)
        {
            XmlElement SCrootNode = doc.CreateElement("SigningCertificate", XmlXAdESTSNamespaceUrl);
            XmlElement scNode = doc.CreateElement("Cert", XmlXAdESTSNamespaceUrl);
            SCrootNode.AppendChild(scNode);
            XmlElement cdNode = doc.CreateElement("CertDigest", XmlXAdESTSNamespaceUrl);
            scNode.AppendChild(cdNode);
            XmlElement scValNode = doc.CreateElement("DigestMethod", XmlXAdESTSNamespaceUrl);
            scValNode.SetAttribute("Algorithm", "http://www.w3.org/2000/09/xmldsig#sha1");
            cdNode.AppendChild(scValNode);
            scValNode = doc.CreateElement("DigestValue", XmlXAdESTSNamespaceUrl);
            scValNode.InnerText = Convert.ToBase64String(cert.GetCertHash());
            cdNode.AppendChild(scValNode);

            cdNode = doc.CreateElement("IssuerSerial", XmlXAdESTSNamespaceUrl);
            scNode.AppendChild(cdNode);
            scValNode = doc.CreateElement("X509IssuerName", XmlDsigNamespaceUrl);
            scValNode.InnerText = cert.IssuerName.Name;
            cdNode.AppendChild(scValNode);
            scValNode = doc.CreateElement("X509SerialNumber", XmlDsigNamespaceUrl);
            scValNode.InnerText = iCryptoLib.HexToDecimal(cert.SerialNumber);
            cdNode.AppendChild(scValNode);

            return SCrootNode;
        }
        /// <summary>
        /// Add a signature property to the document
        /// </summary>
        /// <param name="content">XML contents of the property</param>
        public void AddProperty(XmlElement content)
        {
            if (content == null)
                throw new ArgumentNullException("content");
            if (String.Compare(content.NamespaceURI, XmlDsigNamespaceUrl) == 0)
                throw new InvalidOperationException("Signature properties must not be in the XML Digital Signature namespace");

            signedSigProperties.AppendChild(content);
            return;
        }

        /// <summary>
        /// Get an element refered to by its ID
        /// </summary>
        /// <param name="doc">XML document to search for the element in</param>
        /// <param name="id">element to search for</param>
        /// <returns>Element with the given ID, null if it could not be found</returns>
        ///

        public override XmlElement GetIdElement(XmlDocument doc, string id)
        {
            if (id == null)
                return null;

            // see if this is the signature properties being referenced, otherwise fall back to default behavior
            if (String.Compare(id, signedProperties.GetAttribute("Id"), StringComparison.OrdinalIgnoreCase) == 0)
                return signedProperties;
            else
                return base.GetIdElement(doc, id);
        }

    }
}